@@ -7,19 +7,22 @@ Trimesh, Scene, PointCloud, and Path objects.
 
 Works on all major platforms: Windows, Linux, and OSX.
 """
+import platform
+import collections
+import numpy as np
 
 import pyglet
 import pyglet.gl as gl
 
-import numpy as np
-
-import platform
-import collections
+from .trackball import Trackball
 
+from .. import util
 from .. import rendering
-from ..transformations import Arcball
-from ..util import log, BytesIO
+
 from ..visual import to_rgba
+from ..transformations import translation_matrix
+
+pyglet.options['shadow_window'] = False
 
 # smooth only when fewer faces than this
 _SMOOTH_MAX_FACES = 100000
@@ -32,12 +35,17 @@ class SceneViewer(pyglet.window.Window):
                  smooth=True,
                  flags=None,
                  visible=True,
-                 resolution=(640, 480),
+                 resolution=None,
                  start_loop=True,
                  callback=None,
                  callback_period=None,
                  caption=None,
-                 **kwargs):
+                 fixed=None,
+                 offset_lines=True,
+                 background=None,
+                 window_conf=None,
+                 profile=False,
+                 ** kwargs):
         """
         Create a window that will display a trimesh.Scene object
         in an OpenGL context via pyglet.
@@ -62,63 +70,96 @@ class SceneViewer(pyglet.window.Window):
           update things in the scene
         callback_period : float
           How often to call the callback, in seconds
+        fixed : None or iterable
+          List of keys in scene.geometry to skip view
+          transform on to keep fixed relative to camera
+        offset_lines : bool
+          If True, will offset lines slightly so if drawn
+          coplanar with mesh geometry they will be visible
+        background : None or (4,) uint8
+          Color for background
+        window_conf : None, or gl.Config
+          Passed to window init
         kwargs : dict
           Additional arguments to pass, including
           'background' for to set background color
         """
         self.scene = self._scene = scene
+
         self.callback = callback
         self.callback_period = callback_period
         self.scene._redraw = self._redraw
+        self.offset_lines = bool(offset_lines)
+        self.background = background
+        # save initial camera transform
+        self._initial_camera_transform = scene.camera_transform.copy()
+
+        # a transform to offset lines slightly to avoid Z-fighting
+        if self.offset_lines:
+            self._line_offset = translation_matrix(
+                [0, 0, scene.scale / 1000])
+
         self.reset_view(flags=flags)
         self.batch = pyglet.graphics.Batch()
+        self._smooth = smooth
+
+        self._profile = bool(profile)
+        if self._profile:
+            from pyinstrument import Profiler
+            self.Profiler = Profiler
 
         # store kwargs
         self.kwargs = kwargs
 
         # store a vertexlist for an axis marker
         self._axis = None
+        # store a vertexlist for a grid display
+        self._grid = None
         # store scene geometry as vertex lists
         self.vertex_list = {}
         # store geometry hashes
         self.vertex_list_hash = {}
         # store geometry rendering mode
         self.vertex_list_mode = {}
-
+        # store meshes that don't rotate relative to viewer
+        self.fixed = fixed
+        # store a hidden (don't not display) node.
+        self._nodes_hidden = set()
         # name : texture
         self.textures = {}
 
-        if scene.camera is not None:
-            if resolution is not None and not np.allclose(
-                    resolution,
-                    scene.camera.resolution,
-                    rtol=0, atol=0):
-                log.warning(
-                    'resolution is overwritten by camera: '
-                    '{} -> {}'.format(resolution,
-                                      scene.camera.resolution))
-                resolution = scene.camera.resolution
+        # if resolution isn't defined set a default value
+        if resolution is None:
+            resolution = scene.camera.resolution
+        else:
+            scene.camera.resolution = resolution
 
-        if 'camera' not in scene.graph:
-            # if the camera hasn't been set, set it now
-            scene.set_camera()
-
-        try:
-            # try enabling antialiasing
-            # if you have a graphics card this will probably work
-            conf = gl.Config(sample_buffers=1,
-                             samples=4,
-                             depth_size=24,
-                             double_buffer=True)
-            super(SceneViewer, self).__init__(config=conf,
-                                              visible=visible,
-                                              resizable=True,
-                                              width=resolution[0],
-                                              height=resolution[1],
-                                              caption=caption)
-        except pyglet.window.NoSuchConfigException:
-            conf = gl.Config(double_buffer=True)
-            super(SceneViewer, self).__init__(config=conf,
+        # no window conf was passed so try to get the best looking one
+        if window_conf is None:
+            try:
+                # try enabling antialiasing
+                # if you have a graphics card this will probably work
+                conf = gl.Config(sample_buffers=1,
+                                 samples=4,
+                                 depth_size=24,
+                                 double_buffer=True)
+                super(SceneViewer, self).__init__(config=conf,
+                                                  visible=visible,
+                                                  resizable=True,
+                                                  width=resolution[0],
+                                                  height=resolution[1],
+                                                  caption=caption)
+            except pyglet.window.NoSuchConfigException:
+                conf = gl.Config(double_buffer=True)
+                super(SceneViewer, self).__init__(config=conf,
+                                                  resizable=True,
+                                                  visible=visible,
+                                                  width=resolution[0],
+                                                  height=resolution[1],
+                                                  caption=caption)
+        else:
+            # window config was manually passed
+            super(SceneViewer, self).__init__(config=window_conf,
                                               resizable=True,
                                               visible=visible,
                                               width=resolution[0],
@@ -126,10 +167,7 @@ class SceneViewer(pyglet.window.Window):
                                               caption=caption)
 
         # add scene geometry to viewer geometry
-        for name, mesh in scene.geometry.items():
-            self.add_geometry(name=name,
-                              geometry=mesh,
-                              smooth=bool(smooth))
+        self._update_vertex_list()
 
         # call after geometry is added
         self.init_gl()
@@ -152,13 +190,23 @@ class SceneViewer(pyglet.window.Window):
     def _redraw(self):
         self.on_draw()
 
+    def _update_vertex_list(self):
+        # update vertex_list if needed
+        for name, geom in self.scene.geometry.items():
+            if geom.is_empty:
+                continue
+            if geometry_hash(geom) == self.vertex_list_hash.get(name):
+                continue
+            self.add_geometry(name=name,
+                              geometry=geom,
+                              smooth=bool(self._smooth))
+
     def _update_meshes(self):
         # call the callback if specified
         if self.callback is not None:
             self.callback(self.scene)
-        for name, mesh in self.scene.geometry.items():
-            if self.vertex_list_hash.get(name, None) != geometry_hash(mesh):
-                self.add_geometry(name, mesh)
+            self._update_vertex_list()
+            self._update_perspective(self.width, self.height)
 
     def add_geometry(self, name, geometry, **kwargs):
         """
@@ -182,12 +230,62 @@ class SceneViewer(pyglet.window.Window):
         # save the rendering mode from the constructor args
         self.vertex_list_mode[name] = args[1]
 
-        # if a geometry has a texture defined convert it to opengl form and
-        # save
-        if hasattr(geometry, 'visual') and hasattr(
-                geometry.visual, 'material'):
-            self.textures[name] = rendering.material_to_texture(
+        try:
+            # if a geometry has UV coordinates that match vertices
+            assert len(geometry.visual.uv) == len(geometry.vertices)
+            has_tex = True
+        except BaseException:
+            has_tex = False
+
+        if has_tex:
+            tex = rendering.material_to_texture(
                 geometry.visual.material)
+            if tex is not None:
+                self.textures[name] = tex
+
+    def cleanup_geometries(self):
+        """
+        Remove any stored vertex lists that no longer
+        exist in the scene.
+        """
+        # shorthand to scene graph
+        graph = self.scene.graph
+        # which parts of the graph still have geometry
+        geom_keep = set([graph[node][1] for
+                         node in graph.nodes_geometry])
+        # which geometries no longer need to be kept
+        geom_delete = [geom for geom in self.vertex_list
+                       if geom not in geom_keep]
+        for geom in geom_delete:
+            # remove stored vertex references
+            self.vertex_list.pop(geom, None)
+            self.vertex_list_hash.pop(geom, None)
+            self.vertex_list_mode.pop(geom, None)
+            self.textures.pop(geom, None)
+
+    def unhide_geometry(self, node):
+        """
+        If a node is hidden remove the flag and show the
+        geometry on the next draw.
+
+        Parameters
+        -------------
+        node : str
+          Node to display
+        """
+        self._nodes_hidden.discard(node)
+
+    def hide_geometry(self, node):
+        """
+        Don't display the geometry contained at a node on
+        the next draw.
+
+        Parameters
+        -------------
+        node : str
+          Node to not display
+        """
+        self._nodes_hidden.add(node)
 
     def reset_view(self, flags=None):
         """
@@ -199,21 +297,18 @@ class SceneViewer(pyglet.window.Window):
           If any view key passed override the default
           e.g. {'cull': False}
         """
-        self.view = {'cull': True,
-                     'axis': False,
-                     'fullscreen': False,
-                     'wireframe': False,
-                     'translation': np.zeros(3),
-                     'center': self.scene.centroid,
-                     'scale': self.scene.scale,
-                     'ball': Arcball()}
-
+        self.view = {
+            'cull': True,
+            'axis': False,
+            'grid': False,
+            'fullscreen': False,
+            'wireframe': False,
+            'ball': Trackball(
+                pose=self._initial_camera_transform,
+                size=self.scene.camera.resolution,
+                scale=self.scene.scale,
+                target=self.scene.centroid)}
         try:
-            # place the arcball (rotation widget) in the center of the view
-            self.view['ball'].place([self.width / 2.0,
-                                     self.height / 2.0],
-                                    (self.width + self.height) / 2.0)
-
             # if any flags are passed override defaults
             if isinstance(flags, dict):
                 for k, v in flags.items():
@@ -225,27 +320,44 @@ class SceneViewer(pyglet.window.Window):
 
     def init_gl(self):
         """
-        Perform the magic incantations to create an OpenGL scene.
+        Perform the magic incantations to create an
+        OpenGL scene using pyglet.
         """
 
-        # default background color is white-ish
-        background = [.99, .99, .99, 1.0]
         # if user passed a background color use it
-        if 'background' in self.kwargs:
-            try:
-                # convert to (4,) uint8 RGBA
-                background = to_rgba(self.kwargs['background'])
-                # convert to 0.0 - 1.0 float
-                background = background.astype(np.float64) / 255.0
-            except BaseException:
-                log.error('background color wrong!',
-                          exc_info=True)
-        # apply the background color
+        if self.background is None:
+            # default background color is white
+            background = np.ones(4)
+        else:
+            # convert to (4,) uint8 RGBA
+            background = to_rgba(self.background)
+            # convert to 0.0-1.0 float
+            background = background.astype(np.float64) / 255.0
+
+        self._gl_set_background(background)
+        # use camera setting for depth
+        self._gl_enable_depth(self.scene.camera)
+        self._gl_enable_color_material()
+        self._gl_enable_blending()
+        self._gl_enable_smooth_lines()
+        self._gl_enable_lighting(self.scene)
+
+    @staticmethod
+    def _gl_set_background(background):
         gl.glClearColor(*background)
 
-        max_depth = (np.abs(self.scene.bounds).max(axis=1) ** 2).sum() ** .5
-        max_depth = np.clip(max_depth, 500.00, np.inf)
-        gl.glDepthRange(0.0, max_depth)
+    @staticmethod
+    def _gl_unset_background():
+        gl.glClearColor(*[0, 0, 0, 0])
+
+    @staticmethod
+    def _gl_enable_depth(camera):
+        """
+        Enable depth test in OpenGL using distances
+        from `scene.camera`.
+        """
+        # set the culling depth from our camera object
+        gl.glDepthRange(camera.z_near, camera.z_far)
 
         gl.glClearDepth(1.0)
         gl.glEnable(gl.GL_DEPTH_TEST)
@@ -253,61 +365,92 @@ class SceneViewer(pyglet.window.Window):
 
         gl.glEnable(gl.GL_DEPTH_TEST)
         gl.glEnable(gl.GL_CULL_FACE)
-        gl.glEnable(gl.GL_LIGHTING)
-        gl.glEnable(gl.GL_LIGHT0)
-        gl.glEnable(gl.GL_LIGHT1)
 
-        # put the light at one corner of the scenes AABB
-        gl.glLightfv(gl.GL_LIGHT0,
-                     gl.GL_POSITION,
-                     rendering.vector_to_gl(np.append(self.scene.bounds[1], 0)))
-        gl.glLightfv(gl.GL_LIGHT0, gl.GL_SPECULAR,
-                     rendering.vector_to_gl(.5, .5, 1, 1))
-        gl.glLightfv(gl.GL_LIGHT0, gl.GL_DIFFUSE,
-                     rendering.vector_to_gl(1, 1, 1, .75))
-        gl.glLightfv(gl.GL_LIGHT0, gl.GL_AMBIENT,
-                     rendering.vector_to_gl(.1, .1, .1, .2))
-
-        gl.glColorMaterial(gl.GL_FRONT_AND_BACK, gl.GL_AMBIENT_AND_DIFFUSE)
+    @staticmethod
+    def _gl_enable_color_material():
+        # do some openGL things
+        gl.glColorMaterial(gl.GL_FRONT_AND_BACK,
+                           gl.GL_AMBIENT_AND_DIFFUSE)
         gl.glEnable(gl.GL_COLOR_MATERIAL)
         gl.glShadeModel(gl.GL_SMOOTH)
 
         gl.glMaterialfv(gl.GL_FRONT,
                         gl.GL_AMBIENT,
-                        rendering.vector_to_gl(0.192250, 0.192250, 0.192250))
+                        rendering.vector_to_gl(
+                            0.192250, 0.192250, 0.192250))
         gl.glMaterialfv(gl.GL_FRONT,
                         gl.GL_DIFFUSE,
-                        rendering.vector_to_gl(0.507540, 0.507540, 0.507540))
+                        rendering.vector_to_gl(
+                            0.507540, 0.507540, 0.507540))
         gl.glMaterialfv(gl.GL_FRONT,
                         gl.GL_SPECULAR,
-                        rendering.vector_to_gl(.5082730, .5082730, .5082730))
+                        rendering.vector_to_gl(
+                            .5082730, .5082730, .5082730))
 
         gl.glMaterialf(gl.GL_FRONT,
                        gl.GL_SHININESS,
                        .4 * 128.0)
 
+    @staticmethod
+    def _gl_enable_blending():
+        # enable blending for transparency
         gl.glEnable(gl.GL_BLEND)
-        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
+        gl.glBlendFunc(gl.GL_SRC_ALPHA,
+                       gl.GL_ONE_MINUS_SRC_ALPHA)
 
+    @staticmethod
+    def _gl_enable_smooth_lines():
+        # make the lines from Path3D objects less ugly
         gl.glEnable(gl.GL_LINE_SMOOTH)
         gl.glHint(gl.GL_LINE_SMOOTH_HINT, gl.GL_NICEST)
-
-        gl.glLineWidth(1.5)
+        # set the width of lines to 4 pixels
+        gl.glLineWidth(4)
+        # set PointCloud markers to 4 pixels in size
         gl.glPointSize(4)
 
+    @staticmethod
+    def _gl_enable_lighting(scene):
+        """
+        Take the lights defined in scene.lights and
+        apply them as openGL lights.
+        """
+        gl.glEnable(gl.GL_LIGHTING)
+        # opengl only supports 7 lights?
+        for i, light in enumerate(scene.lights[:7]):
+            # the index of which light we have
+            lightN = eval('gl.GL_LIGHT{}'.format(i))
+
+            # get the transform for the light by name
+            matrix = scene.graph.get(light.name)[0]
+
+            # convert light object to glLightfv calls
+            multiargs = rendering.light_to_gl(
+                light=light,
+                transform=matrix,
+                lightN=lightN)
+
+            # enable the light in question
+            gl.glEnable(lightN)
+            # run the glLightfv calls
+            for args in multiargs:
+                gl.glLightfv(*args)
+
     def toggle_culling(self):
         """
-        Toggle backface culling on or off. It is on by default
-        but if you are dealing with non- watertight meshes you
-        probably want to be able to see the back sides.
+        Toggle back face culling.
+
+        It is on by default but if you are dealing with
+        non- watertight meshes you probably want to be able
+        to see the back sides.
         """
         self.view['cull'] = not self.view['cull']
         self.update_flags()
 
     def toggle_wireframe(self):
         """
-        Toggle unfilled wireframe mode on or off, good for
-        looking inside meshes. Off by default.
+        Toggle wireframe mode
+
+        Good for  looking inside meshes, off by default.
         """
         self.view['wireframe'] = not self.view['wireframe']
         self.update_flags()
@@ -321,8 +464,8 @@ class SceneViewer(pyglet.window.Window):
 
     def toggle_axis(self):
         """
-        Toggle a rendered XYZ/RGB axis marker on, world frame,
-        or every frame. Off by default.
+        Toggle a rendered XYZ/RGB axis marker:
+        off, world frame, every frame
         """
         # cycle through three axis states
         states = [False, 'world', 'all']
@@ -333,10 +476,18 @@ class SceneViewer(pyglet.window.Window):
         # perform gl actions
         self.update_flags()
 
+    def toggle_grid(self):
+        """
+        Toggle a rendered grid.
+        """
+        # update state to next index
+        self.view['grid'] = not self.view['grid']
+        # perform gl actions
+        self.update_flags()
+
     def update_flags(self):
         """
-        Check the view flags and call what is needed with gl
-        to handle it correctly.
+        Check the view flags, and call required GL functions.
         """
         # view mode, filled vs wirefrom
         if self.view['wireframe']:
@@ -363,7 +514,6 @@ class SceneViewer(pyglet.window.Window):
             args = rendering.mesh_to_vertexlist(axis)
             # store the axis as a reference
             self._axis = self.batch.add_indexed(*args)
-
         # case where we DON'T want an axis but a vertexlist
         # IS stored internally
         elif not self.view['axis'] and self._axis is not None:
@@ -372,7 +522,34 @@ class SceneViewer(pyglet.window.Window):
             # set the reference to None
             self._axis = None
 
-    def on_resize(self, width, height):
+        if self.view['grid'] and self._grid is None:
+            try:
+                # create a grid marker
+                from ..path.creation import grid
+                bounds = self.scene.bounds
+                center = bounds.mean(axis=0)
+                # set the grid to the lowest Z position
+                # also offset by the scale to avoid interference
+                center[2] = bounds[0][2] - (bounds[:, 2].ptp() / 100)
+                # choose the side length by maximum XY length
+                side = bounds.ptp(axis=0)[:2].max()
+                # create an axis marker sized relative to the scene
+                grid_mesh = grid(
+                    side=side,
+                    count=4,
+                    transform=translation_matrix(center))
+                # convert the path to vertexlist args
+                args = rendering.convert_to_vertexlist(grid_mesh)
+                # create ordered args for a vertex list
+                self._grid = self.batch.add_indexed(*args)
+            except BaseException:
+                util.log.warning(
+                    'failed to create grid!', exc_info=True)
+        elif not self.view['grid'] and self._grid is not None:
+            self._grid.delete()
+            self._grid = None
+
+    def _update_perspective(self, width, height):
         try:
             # for high DPI screens viewport size
             # will be different then the passed size
@@ -386,37 +563,67 @@ class SceneViewer(pyglet.window.Window):
         gl.glMatrixMode(gl.GL_PROJECTION)
         gl.glLoadIdentity()
 
-        if self.scene.camera is None:
-            fovY = 60.0
-        else:
-            fovY = self.scene.camera.fov[1]
+        # get field of view and Z range from camera
+        camera = self.scene.camera
 
-        gl.gluPerspective(fovY,
+        # set perspective from camera data
+        gl.gluPerspective(camera.fov[1],
                           width / float(height),
-                          .01,
-                          self.scene.scale * 5.0)
+                          camera.z_near,
+                          camera.z_far)
         gl.glMatrixMode(gl.GL_MODELVIEW)
-        self.view['ball'].place([width / 2, height / 2],
-                                (width + height) / 2)
+
+        return width, height
+
+    def on_resize(self, width, height):
+        """
+        Handle resized windows.
+        """
+        width, height = self._update_perspective(width, height)
+        self.scene.camera.resolution = (width, height)
+        self.view['ball'].resize(self.scene.camera.resolution)
+        self.scene.camera_transform[...] = self.view['ball'].pose
 
     def on_mouse_press(self, x, y, buttons, modifiers):
-        self.view['ball'].down([x, -y])
+        """
+        Set the start point of the drag.
+        """
+        self.view['ball'].set_state(Trackball.STATE_ROTATE)
+        if (buttons == pyglet.window.mouse.LEFT):
+            ctrl = (modifiers & pyglet.window.key.MOD_CTRL)
+            shift = (modifiers & pyglet.window.key.MOD_SHIFT)
+            if (ctrl and shift):
+                self.view['ball'].set_state(Trackball.STATE_ZOOM)
+            elif shift:
+                self.view['ball'].set_state(Trackball.STATE_ROLL)
+            elif ctrl:
+                self.view['ball'].set_state(Trackball.STATE_PAN)
+        elif (buttons == pyglet.window.mouse.MIDDLE):
+            self.view['ball'].set_state(Trackball.STATE_PAN)
+        elif (buttons == pyglet.window.mouse.RIGHT):
+            self.view['ball'].set_state(Trackball.STATE_ZOOM)
+
+        self.view['ball'].down(np.array([x, y]))
+        self.scene.camera_transform[...] = self.view['ball'].pose
 
     def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
-        # left mouse button, with control key down (pan)
-        if ((buttons == pyglet.window.mouse.LEFT) and
-                (modifiers & pyglet.window.key.MOD_CTRL)):
-            delta = [dx / self.width, dy / self.height]
-            self.view['translation'][0:2] += delta
-
-        # left mouse button, no modifier keys pressed (rotate)
-        elif (buttons == pyglet.window.mouse.LEFT):
-            self.view['ball'].drag([x, -y])
+        """
+        Pan or rotate the view.
+        """
+        self.view['ball'].drag(np.array([x, y]))
+        self.scene.camera_transform[...] = self.view['ball'].pose
 
     def on_mouse_scroll(self, x, y, dx, dy):
-        self.view['translation'][2] += float(dy) / self.height
+        """
+        Zoom the view.
+        """
+        self.view['ball'].scroll(dy)
+        self.scene.camera_transform[...] = self.view['ball'].pose
 
     def on_key_press(self, symbol, modifiers):
+        """
+        Call appropriate functions given key presses.
+        """
         magnitude = 10
         if symbol == pyglet.window.key.W:
             self.toggle_wireframe()
@@ -426,44 +633,54 @@ class SceneViewer(pyglet.window.Window):
             self.toggle_culling()
         elif symbol == pyglet.window.key.A:
             self.toggle_axis()
+        elif symbol == pyglet.window.key.G:
+            self.toggle_grid()
         elif symbol == pyglet.window.key.Q:
-            self.close()
+            self.on_close()
         elif symbol == pyglet.window.key.M:
             self.maximize()
         elif symbol == pyglet.window.key.F:
             self.toggle_fullscreen()
-        elif symbol == pyglet.window.key.LEFT:
+
+        if symbol in [
+                pyglet.window.key.LEFT,
+                pyglet.window.key.RIGHT,
+                pyglet.window.key.DOWN,
+                pyglet.window.key.UP]:
             self.view['ball'].down([0, 0])
-            self.view['ball'].drag([-magnitude, 0])
-        elif symbol == pyglet.window.key.RIGHT:
-            self.view['ball'].down([0, 0])
-            self.view['ball'].drag([magnitude, 0])
-        elif symbol == pyglet.window.key.DOWN:
-            self.view['ball'].down([0, 0])
-            self.view['ball'].drag([0, -magnitude])
-        elif symbol == pyglet.window.key.UP:
-            self.view['ball'].down([0, 0])
-            self.view['ball'].drag([0, magnitude])
+            if symbol == pyglet.window.key.LEFT:
+                self.view['ball'].drag([-magnitude, 0])
+            elif symbol == pyglet.window.key.RIGHT:
+                self.view['ball'].drag([magnitude, 0])
+            elif symbol == pyglet.window.key.DOWN:
+                self.view['ball'].drag([0, -magnitude])
+            elif symbol == pyglet.window.key.UP:
+                self.view['ball'].drag([0, magnitude])
+            self.scene.camera_transform[...] = self.view['ball'].pose
 
     def on_draw(self):
+        """
+        Run the actual draw calls.
+        """
+
+        if self._profile:
+            profiler = self.Profiler()
+            profiler.start()
+
         self._update_meshes()
         gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
         gl.glLoadIdentity()
 
         # pull the new camera transform from the scene
-        transform_camera, _junk = self.scene.graph['camera']
+        transform_camera = np.linalg.inv(self.scene.camera_transform)
 
         # apply the camera transform to the matrix stack
         gl.glMultMatrixf(rendering.matrix_to_gl(transform_camera))
 
-        # dragging the mouse moves the view transform
-        # but doesn't alter the scene
-        transform_view = view_to_transform(self.view)
-        gl.glMultMatrixf(rendering.matrix_to_gl(transform_view))
-
         # we want to render fully opaque objects first,
         # followed by objects which have transparency
         node_names = collections.deque(self.scene.graph.nodes_geometry)
+        # how many nodes did we start with
         count_original = len(node_names)
         count = -1
 
@@ -471,20 +688,48 @@ class SceneViewer(pyglet.window.Window):
         if self._axis:
             # we stored it as a vertex list
             self._axis.draw(mode=gl.GL_TRIANGLES)
+        if self._grid:
+            self._grid.draw(mode=gl.GL_LINES)
 
         while len(node_names) > 0:
             count += 1
             current_node = node_names.popleft()
 
+            if current_node in self._nodes_hidden:
+                continue
+
             # get the transform from world to geometry and mesh name
-            transform, geometry_name = self.scene.graph[current_node]
+            transform, geometry_name = self.scene.graph.get(current_node)
 
             # if no geometry at this frame continue without rendering
             if geometry_name is None:
                 continue
 
+            # if a geometry is marked as fixed apply the inverse view transform
+            if self.fixed is not None and geometry_name in self.fixed:
+                # remove altered camera transform from fixed geometry
+                transform_fix = np.linalg.inv(
+                    np.dot(self._initial_camera_transform, transform_camera))
+                # apply the transform so the fixed geometry doesn't move
+                transform = np.dot(transform, transform_fix)
+
             # get a reference to the mesh so we can check transparency
             mesh = self.scene.geometry[geometry_name]
+            if mesh.is_empty:
+                continue
+
+            # get the GL mode of the current geometry
+            mode = self.vertex_list_mode[geometry_name]
+
+            # if you draw a coplanar line with a triangle it will z-fight
+            # the best way to do this is probably a shader but this works fine
+            if mode == gl.GL_LINES:
+                # apply the offset in camera space
+                transform = util.multi_dot([
+                    transform,
+                    np.linalg.inv(transform_camera),
+                    self._line_offset,
+                    transform_camera])
 
             # add a new matrix to the model stack
             gl.glPushMatrix()
@@ -508,23 +753,26 @@ class SceneViewer(pyglet.window.Window):
                     # come back to this mesh later
                     continue
 
-            #
+            # if we have texture enable the target texture
             texture = None
             if geometry_name in self.textures:
                 texture = self.textures[geometry_name]
                 gl.glEnable(texture.target)
                 gl.glBindTexture(texture.target, texture.id)
 
-            # get the mode of the current geometry
-            mode = self.vertex_list_mode[geometry_name]
             # draw the mesh with its transform applied
             self.vertex_list[geometry_name].draw(mode=mode)
             # pop the matrix stack as we drew what we needed to draw
             gl.glPopMatrix()
 
+            # disable texture after using
             if texture is not None:
                 gl.glDisable(texture.target)
 
+        if self._profile:
+            profiler.stop()
+            print(profiler.output_text(unicode=True, color=True))
+
     def save_image(self, file_obj):
         """
         Save the current color buffer to a file object
@@ -544,18 +792,6 @@ class SceneViewer(pyglet.window.Window):
             colorbuffer.save(filename=file_obj)
 
 
-def view_to_transform(view):
-    """
-    Given a dictionary containing view parameters,
-    calculate a transformation matrix.
-    """
-    transform = view['ball'].matrix()
-    transform[0:3, 3] = view['center']
-    transform[0:3, 3] -= np.dot(transform[0:3, 0:3], view['center'])
-    transform[0:3, 3] += view['translation'] * view['scale'] * 5.0
-    return transform
-
-
 def geometry_hash(geometry):
     """
     Get an MD5 for a geometry object
@@ -582,7 +818,7 @@ def geometry_hash(geometry):
 
 
 def render_scene(scene,
-                 resolution=(1080, 1080),
+                 resolution=None,
                  visible=True,
                  **kwargs):
     """
@@ -592,8 +828,8 @@ def render_scene(scene,
     ------------
     scene : trimesh.Scene
       Geometry to be rendered
-    resolution : (2,) int
-      Resolution in pixels
+    resolution : (2,) int or None
+      Resolution in pixels, or set from scene.camera
     kwargs : **
       Passed to SceneViewer
 
@@ -611,19 +847,21 @@ def render_scene(scene,
     if visible is None:
         visible = platform.system() != 'Linux'
 
+    from ..util import BytesIO
+
     # need to run loop twice to display anything
-    for i in range(2):
+    for save in [False, False, True]:
         pyglet.clock.tick()
         window.switch_to()
         window.dispatch_events()
         window.dispatch_event('on_draw')
         window.flip()
-
-    # save the color buffer data to memory
-    file_obj = BytesIO()
-    window.save_image(file_obj)
-    file_obj.seek(0)
-    render = file_obj.read()
+        if save:
+            # save the color buffer data to memory
+            file_obj = BytesIO()
+            window.save_image(file_obj)
+            file_obj.seek(0)
+            render = file_obj.read()
     window.close()
 
     return render
